@@ -82,6 +82,176 @@ typedef struct {
 #define COL_AUTH_PAGE 1
 #define COL_AUTH_TYPE 2
 
+static NMConnection *
+import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
+{
+	NMConnection *connection;
+	NMSettingConnection *s_con;
+	NMSettingVPN *s_vpn;
+	NMSettingIP4Config *s_ip4;
+	GKeyFile *keyfile;
+	GKeyFileFlags flags;
+	const char *buf;
+	gboolean bval;
+
+	keyfile = g_key_file_new ();
+	flags = G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS;
+
+	if (!g_key_file_load_from_file (keyfile, path, flags, error)) {
+		g_set_error (error, 0, 0, "does not look like a %s VPN connection (parse failed)", OPENCONNECT_PLUGIN_NAME);
+		return NULL;
+	}
+
+	connection = nm_connection_new ();
+	s_con = NM_SETTING_CONNECTION (nm_setting_connection_new ());
+	nm_connection_add_setting (connection, NM_SETTING (s_con));
+
+	s_vpn = NM_SETTING_VPN (nm_setting_vpn_new ());
+	g_object_set (s_vpn, NM_SETTING_VPN_SERVICE_TYPE, NM_DBUS_SERVICE_OPENCONNECT, NULL);
+	nm_connection_add_setting (connection, NM_SETTING (s_vpn));
+
+	s_ip4 = NM_SETTING_IP4_CONFIG (nm_setting_ip4_config_new ());
+	nm_connection_add_setting (connection, NM_SETTING (s_ip4));
+
+	/* Host */
+	buf = g_key_file_get_string (keyfile, "openconnect", "Host", NULL);
+	if (buf) {
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_GATEWAY, buf);
+	} else {
+		g_set_error (error, 0, 0, "does not look like a %s VPN connection (no Host)", OPENCONNECT_PLUGIN_NAME);
+		g_object_unref (connection);
+		return NULL;
+	}
+
+	/* Optional Settings */
+
+	/* Description */
+	buf = g_key_file_get_string (keyfile, "openconnect", "Description", NULL);
+	if (buf)
+		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);
+
+	/* CA Certificate */
+	buf = g_key_file_get_string (keyfile, "openconnect", "CACert", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CACERT, buf);
+
+	/* Proxy */
+	buf = g_key_file_get_string (keyfile, "openconnect", "Proxy", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY, buf);
+
+	/* Cisco Secure Desktop */
+	bval = g_key_file_get_boolean (keyfile, "openconnect", "CSDEnable", NULL);
+	if (bval)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE, "yes");
+
+	/* User Certificate */
+	buf = g_key_file_get_string (keyfile, "openconnect", "UserCertificate", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT, buf);
+
+	/* Private Key */
+	buf = g_key_file_get_string (keyfile, "openconnect", "PrivateKey", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PRIVKEY, buf);
+
+	/* FSID */
+	bval = g_key_file_get_boolean (keyfile, "openconnect", "FSID", NULL);
+	if (bval)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, "yes");
+
+	return connection;
+}
+
+static gboolean
+export (NMVpnPluginUiInterface *iface,
+        const char *path,
+        NMConnection *connection,
+        GError **error)
+{
+	NMSettingConnection *s_con;
+	NMSettingIP4Config *s_ip4;
+	NMSettingVPN *s_vpn;
+	const char *value;
+	const char *gateway = NULL;
+	const char *cacert = NULL;
+	const char *proxy = NULL;
+	gboolean csd_enable = FALSE;
+	const char *usercert = NULL;
+	const char *privkey = NULL;
+	gboolean pem_passphrase_fsid = FALSE;
+	gboolean success = FALSE;
+	FILE *f;
+
+	f = fopen (path, "w");
+	if (!f) {
+		g_set_error (error, 0, 0, "could not open file for writing");
+		return FALSE;
+	}
+
+	s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
+	s_ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
+
+	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_GATEWAY);
+	if (value && strlen (value))
+		gateway = value;
+	else {
+		g_set_error (error, 0, 0, "connection was incomplete (missing gateway)");
+		goto done;
+	}
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CACERT);
+	if (value && strlen (value))
+		cacert = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY);
+	if (value && strlen (value))
+		proxy = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE);
+	if (value && !strcmp (value, "yes"))
+		csd_enable = TRUE;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT);
+	if (value && strlen (value))
+		usercert = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PRIVKEY);
+	if (value && strlen (value))
+		privkey = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
+	if (value && !strcmp (value, "yes"))
+		pem_passphrase_fsid = TRUE;
+
+	fprintf (f,
+		 "[openconnect]\n"
+		 "Description=%s\n"
+		 "Host=%s\n"
+		 "CACert=%s\n"
+		 "Proxy=%s\n"
+		 "CSDEnable=%s\n"
+		 "UserCertificate=%s\n"
+		 "PrivateKey=%s\n"
+		 "FSID=%s\n",
+		 /* Description */           nm_setting_connection_get_id (s_con),
+		 /* Host */                  gateway,
+		 /* CA Certificate */        cacert,
+		 /* Proxy */                 proxy ? proxy : "",
+		 /* Cisco Secure Desktop */  csd_enable ? "1" : "0",
+		 /* User Certificate */      usercert,
+		 /* Private Key */           privkey,
+		 /* FSID */                  pem_passphrase_fsid ? "1" : "0");
+
+	success = TRUE;
+
+done:
+	fclose (f);
+	return success;
+}
+
 GQuark
 openconnect_plugin_ui_error_quark (void)
 {
@@ -122,9 +292,6 @@ check_validity (OpenconnectPluginUiWidget *self, GError **error)
 	OpenconnectPluginUiWidgetPrivate *priv = OPENCONNECT_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
 	GtkWidget *widget;
 	const char *str;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
-	const char *contype = NULL;
 
 	widget = glade_xml_get_widget (priv->xml, "gateway_entry");
 	str = gtk_entry_get_text (GTK_ENTRY (widget));
@@ -136,13 +303,19 @@ check_validity (OpenconnectPluginUiWidget *self, GError **error)
 		return FALSE;
 	}
 
-	widget = glade_xml_get_widget (priv->xml, "auth_combo");
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
-	g_assert (model);
-	g_assert (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter));
 
-	gtk_tree_model_get (model, &iter, COL_AUTH_TYPE, &contype, -1);
-	if (!auth_widget_check_validity (priv->xml, contype, error))
+	widget = glade_xml_get_widget (priv->xml, "proxy_entry");
+	str = gtk_entry_get_text (GTK_ENTRY (widget));
+	if (str && str[0] &&
+		strncmp(str, "socks://", 8) && strncmp(str, "http://", 7)) {
+		g_set_error (error,
+		             OPENCONNECT_PLUGIN_UI_ERROR,
+		             OPENCONNECT_PLUGIN_UI_ERROR_INVALID_PROPERTY,
+		             NM_OPENCONNECT_KEY_PROXY);
+		return FALSE;
+	}
+
+	if (!auth_widget_check_validity (priv->xml, error))
 		return FALSE;
 
 	return TRUE;
@@ -154,40 +327,13 @@ stuff_changed_cb (GtkWidget *widget, gpointer user_data)
 	g_signal_emit_by_name (OPENCONNECT_PLUGIN_UI_WIDGET (user_data), "changed");
 }
 
-static void
-auth_combo_changed_cb (GtkWidget *combo, gpointer user_data)
-{
-	OpenconnectPluginUiWidget *self = OPENCONNECT_PLUGIN_UI_WIDGET (user_data);
-	OpenconnectPluginUiWidgetPrivate *priv = OPENCONNECT_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
-	GtkWidget *auth_notebook;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
-	gint new_page = 0;
-
-	auth_notebook = glade_xml_get_widget (priv->xml, "auth_notebook");
-	g_assert (auth_notebook);
-
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
-	g_assert (model);
-	g_assert (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (combo), &iter));
-
-	gtk_tree_model_get (model, &iter, COL_AUTH_PAGE, &new_page, -1);
-	gtk_notebook_set_current_page (GTK_NOTEBOOK (auth_notebook), new_page);
-
-	stuff_changed_cb (combo, self);
-}
-
 static gboolean
 init_plugin_ui (OpenconnectPluginUiWidget *self, NMConnection *connection, GError **error)
 {
 	OpenconnectPluginUiWidgetPrivate *priv = OPENCONNECT_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
 	NMSettingVPN *s_vpn;
 	GtkWidget *widget;
-	GtkListStore *store;
-	GtkTreeIter iter;
-	int active = -1;
 	const char *value;
-	const char *contype = NM_OPENCONNECT_AUTHTYPE_PASSWORD;
 
 	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
 
@@ -204,64 +350,40 @@ init_plugin_ui (OpenconnectPluginUiWidget *self, NMConnection *connection, GErro
 	}
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
 
-	widget = glade_xml_get_widget (priv->xml, "auth_combo");
+	widget = glade_xml_get_widget (priv->xml, "proxy_entry");
 	if (!widget)
 		return FALSE;
 	gtk_size_group_add_widget (priv->group, widget);
-
-	store = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);
-
 	if (s_vpn) {
-		contype = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_AUTHTYPE);
-		if (contype) {
-			if (   strcmp (contype, NM_OPENCONNECT_AUTHTYPE_CERT)
-			    && strcmp (contype, NM_OPENCONNECT_AUTHTYPE_CERT_TPM)
-			    && strcmp (contype, NM_OPENCONNECT_AUTHTYPE_PASSWORD))
-				contype = NM_OPENCONNECT_AUTHTYPE_PASSWORD;
-		} else
-			contype = NM_OPENCONNECT_AUTHTYPE_PASSWORD;
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY);
+		if (value)
+			gtk_entry_set_text (GTK_ENTRY (widget), value);
 	}
+	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
+
+	widget = glade_xml_get_widget (priv->xml, "fsid_button");
+	if (!widget)
+		return FALSE;
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
+		if (value && !strcmp(value, "yes"))
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (widget), TRUE);
+	}
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (stuff_changed_cb), self);
+
+	widget = glade_xml_get_widget (priv->xml, "csd_button");
+	if (!widget)
+		return FALSE;
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE);
+		if (value && !strcmp(value, "yes"))
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (widget), TRUE);
+	}
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (stuff_changed_cb), self);
 
-	/* SecurID/password auth widget */
 	tls_pw_init_auth_widget (priv->xml, priv->group, s_vpn,
-							 NM_OPENCONNECT_AUTHTYPE_PASSWORD, "pw",
 							 stuff_changed_cb, self);
 
-	gtk_list_store_append (store, &iter);
-	gtk_list_store_set (store, &iter,
-	                    COL_AUTH_NAME, _("Password / SecurID"),
-	                    COL_AUTH_PAGE, 1,
-	                    COL_AUTH_TYPE, NM_OPENCONNECT_AUTHTYPE_PASSWORD,
-	                    -1);
-
-	/* Certificate auth widget */
-	tls_pw_init_auth_widget (priv->xml, priv->group, s_vpn,
-							 NM_OPENCONNECT_AUTHTYPE_CERT, "cert",
-							 stuff_changed_cb, self);
-
-	gtk_list_store_append (store, &iter);
-	gtk_list_store_set (store, &iter,
-	                    COL_AUTH_NAME, _("Certificate (TLS)"),
-	                    COL_AUTH_PAGE, 0,
-	                    COL_AUTH_TYPE, NM_OPENCONNECT_AUTHTYPE_CERT,
-	                    -1);
-	if ((active < 0) && !strcmp (contype, NM_OPENCONNECT_AUTHTYPE_CERT))
-		active = 1;
-
-	gtk_list_store_append (store, &iter);
-	gtk_list_store_set (store, &iter,
-	                    COL_AUTH_NAME, _("Certificate (TLS) with TPM"),
-	                    COL_AUTH_PAGE, 0,
-	                    COL_AUTH_TYPE, NM_OPENCONNECT_AUTHTYPE_CERT_TPM,
-	                    -1);
-	if ((active < 0) && !strcmp (contype, NM_OPENCONNECT_AUTHTYPE_CERT_TPM))
-		active = 2;
-
-	gtk_combo_box_set_model (GTK_COMBO_BOX (widget), GTK_TREE_MODEL (store));
-	g_object_unref (store);
-	g_signal_connect (widget, "changed", G_CALLBACK (auth_combo_changed_cb), self);
-	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), active < 0 ? 0 : active);
-
 	return TRUE;
 }
 
@@ -284,8 +406,6 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 	NMSettingVPN *s_vpn;
 	GtkWidget *widget;
 	char *str;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
 	const char *auth_type = NULL;
 
 	if (!check_validity (self, error))
@@ -294,19 +414,25 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 	s_vpn = NM_SETTING_VPN (nm_setting_vpn_new ());
 	g_object_set (s_vpn, NM_SETTING_VPN_SERVICE_TYPE, NM_DBUS_SERVICE_OPENCONNECT, NULL);
 
-	/* Gateway */
 	widget = glade_xml_get_widget (priv->xml, "gateway_entry");
 	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
 	if (str && strlen (str))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_GATEWAY, str);
 
-	widget = glade_xml_get_widget (priv->xml, "auth_combo");
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
-	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter)) {
-		gtk_tree_model_get (model, &iter, COL_AUTH_TYPE, &auth_type, -1);
-		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_AUTHTYPE, auth_type);
-		auth_widget_update_connection (priv->xml, auth_type, s_vpn);
-	}
+	widget = glade_xml_get_widget (priv->xml, "proxy_entry");
+	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
+	if (str && strlen (str))
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY, str);
+
+	widget = glade_xml_get_widget (priv->xml, "fsid_button");
+	str = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (widget))?"yes":"no";
+	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, str);
+
+	widget = glade_xml_get_widget (priv->xml, "csd_button");
+	str = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (widget))?"yes":"no";
+	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE, str);
+	
+	auth_widget_update_connection (priv->xml, auth_type, s_vpn);
 
 	nm_connection_add_setting (connection, NM_SETTING (s_vpn));
 	return TRUE;
@@ -424,7 +550,7 @@ delete_connection (NMVpnPluginUiInterface *iface,
 static guint32
 get_capabilities (NMVpnPluginUiInterface *iface)
 {
-	return 0;
+	return (NM_VPN_PLUGIN_UI_CAPABILITY_IMPORT | NM_VPN_PLUGIN_UI_CAPABILITY_EXPORT);
 }
 
 static NMVpnPluginUiWidgetInterface *
@@ -484,6 +610,8 @@ openconnect_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class)
 	/* interface implementation */
 	iface_class->ui_factory = ui_factory;
 	iface_class->get_capabilities = get_capabilities;
+	iface_class->import_from_file = import;
+	iface_class->export_to_file = export;
 	iface_class->delete_connection = delete_connection;
 }
 
